/*********************************************************************************
*         Beispiel Ansteuerung SAA1064 (plus LM75)
*         ========================================
*                   Uwe Berger; 2009
*
* Temperaturwert aus LM75 auslesen und auf 4 Digits einer 7-Segment-Anzeige
* ausgeben. Kommunikation erfolgt TWI.
*
*  ---------
*  Have fun!
*
**********************************************************************************/
#ifndef F_CPU
	#define F_CPU 12000000
#endif

#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#include "i2cmaster.h"

#define PRELOAD 55000
#define ADDR_LM75 	0x92
#define ADDR_7SEGM 	0x70

// eventuell an eigene Verdrahtung anpassen (SAA1064 -> 7-Segment-Anz.)
// Bit:		7	6	5	4	3	2	1	0
// Segment:	p	g	f	e	d	c	b	a
//
//  -a-
// |f  |b
//  -g-
// |e  |c
//  -d- .p
const uint8_t digit[] PROGMEM = {
	0b00111111,				// 0	
	0b00000110,				// 1
	0b01011011,				// 2
	0b01001111,				// 3
	0b01100110,				// 4
	0b01101101,				// 5
	0b01111101,				// 6
	0b00000111,				// 7
	0b01111111,				// 8
	0b01101111,				// 9
	0b01000000				// Minuszeichen
};

volatile uint8_t click = 0;
volatile uint8_t h_byte;	// Hight-Byte Temperatur-Registers LM75
volatile uint8_t l_byte;	// Low-Byte Temperatur-Registers LM75

//************************************************************************
ISR(TIMER1_OVF_vect)
{
	TCNT1 = PRELOAD;						// Preloadwert setzen
	click = 1;
}

//************************************************************************
void my_delay_ms(uint16_t ms)
{
	uint16_t i;
	for (i=0; i<ms; i++) _delay_ms(1);
}


//************************************************************************
void lm75_init(void)
{
	i2c_start(ADDR_LM75+I2C_WRITE);
	i2c_write(0);							// Funktion Temperatur lesen
	i2c_stop();
	my_delay_ms(100);						// etwas warten bevor es weiter geht...
}


//************************************************************************
void lm75_read(void)
{
	i2c_start(ADDR_LM75+I2C_READ);
	h_byte = i2c_readAck();
	l_byte = i2c_readNak();
	i2c_stop();
}


//************************************************************************
void clear_all_digits(void)
{
	uint8_t i;
	i2c_start(ADDR_7SEGM+I2C_WRITE);
	i2c_write(1);						// auf Adresse 1 "positionieren"
	for (i=1; i<5; i++)	i2c_write(0);	// alle Segmente ausschalten
}


//************************************************************************
void saa1064_init(void)
{
	// SAA1064 Control-Register (Adresse 0)
    // Bit	7	6	5	4	3	2	1	0
	//		|	|	|	|	|	|	|	|
	//		|	|	|	|	|	|	|	- 0: Digit 1,2; 1: Digit 1...4
	//		|	|	|	|	|	|	----- 0/1: Digit 1,3 aus/an
	//		|	|	|	|	|	--------- 0/1: Digit 2,4 aus/an
	//		|	|	|	|	------------- 1: Testmode, alle Segmente an
	//		|	|	|	----------------- 1: plus 3mA
	//		|	|	--------------------- 1: plus 6mA
	//		|	------------------------- 1: plus 12mA
	//		----------------------------- nicht belegt
	i2c_start(ADDR_7SEGM+I2C_WRITE);
	// Controlregister setzen
	i2c_write(0);
	i2c_write(0b01110111);	// Digit 1...4, alle anwaehlbar, 21mA
	clear_all_digits();		// Anzeige loeschen
}

//************************************************************************
// Parameter: 
//     d  -> Digit (1...4)
//     v  -> anzuzeigender Wert (Zahlen 0...9, "-"; -> array digit[]
//     dp -> Punktsegment: 0 nein; >0 ja
void write_digit(uint8_t d, uint8_t v, uint8_t dp)
{
	i2c_start(ADDR_7SEGM+I2C_WRITE);
	i2c_write(d);
	if (dp) {
		i2c_write(pgm_read_byte(&digit[v])+0b10000000);
	} else {
		i2c_write(pgm_read_byte(&digit[v]));
	}
}

//************************************************************************
//************************************************************************
//************************************************************************
int main(void)
{
	uint8_t t;
	
	// Timer initialisieren...
    TCNT1   = PRELOAD;						// Preloadwert setzen
	TCCR1B |= (1 << CS02) | (1 << CS00);	// Prescaler Timer1 auf 1024 stellen
	TIMSK  |= (1 << TOIE1);					// Timer1 Overflow Interrupt enable
	sei(); 									// Interrupts einschalten

	i2c_init();								// TWI initialisieren
	lm75_init();							// LM75 initialisieren	
	saa1064_init();							// SAA1064 initialisieren
	
	while (1){
		if (click) {
			click = 0;
			lm75_read();					// LM75 auslesen
			clear_all_digits();				// Anzeige loeschen
			// Formatierung/Ausgabe Temperatur
			// per Definition: bei negativen Temperaturwerten kann es keine
			// 100ter Stelle geben, da wird dann das Vorzeichen ausgeben...
			if (h_byte >= 128) {			// wenn Bit7 gesetzt, dann negative Temp.
				h_byte = ~(h_byte);			// dann (lt. Datenblatt) Byte invertieren
				h_byte++;					// und plus 1
				//negativer Temperaturwert
				write_digit(1, 10, 0);		// Vorzeichen Minus
			} else {
				//positiver Temperaturwert
				t = h_byte/100;				// 100ter Stelle, nur bei pos. Werten
				if (t) {
					h_byte = h_byte - t*100;
					write_digit(1, t, 0);
				}
			}
			t = h_byte/10;					// 10er Stelle
			if (t) {
				h_byte = h_byte - t*10;
				write_digit(2, t, 0);
			}
			write_digit(3, h_byte, 1);		// 1er Stelle (mit Punkt)
			if (l_byte >= 128) {			// wenn Bit7 gesetzt, dann ,5 sonst ,0
				write_digit(4, 5, 0);
			} else {
				write_digit(4, 0, 0);
			}
		}
	}					
}
